#!/usr/bin/env ruby

#command line
#/Users/hagiwark/Desktop/Ver11/HMonHHAnalysis.rb /Users/hagiwark/Desktop/Ver11/analysisFileA1AT6_Control54_Control200_Control1
#

if ARGV.length != 1
  STDERR.puts "Usage: HMonHHAnalysis.rb analysisFile.txt\nQuitting program..."
  exit
end

$: << File.dirname( __FILE__ ) + "/sub"
require "HGeneRb.rb"
require "yaml"

#Preparation
$rgHashAr = Array.new( 15 ){ | i | Hash.new }
PtPool = 0; CtPool = 1; Extra = 2
Mode_FoldOverlap = -1; Mode_MultipleColor = -2

#Make an instance
hh = HGeneRb.new

hh.readOptionsAndAnalysisFile( ARGV )

if hh.patientPoolFiles != []
  hh.readAnnotationFile
  hh.readGenotypingData( PtPool )
  hh.readGenotypingData( CtPool )
  hh.readGenotypingData( Extra )
  hh.outputSumSNPFile( "#{ hh.outputDirectory }/__sumSNP.txt" )
else
  hh.readSumSNPFile
end
if ( hh.suspectedGenotypingErrorRate != 0.0 ) && ( hh.appliedErrorCorrectionRate == 0.0 )
  hh.correctErrorHM
  hh.outputSumSNPFile( "#{ hh.outputDirectory }/__sumSNP_ErrorCorrectedHM.txt" )
end

#Obtain and output RHSs for each subject.
for i in 0...hh.noPtPool do $rgHashAr[ 1 ][ ( "_%03d" % i ).to_sym ] = hh.obtainRHS_C( hh.rHSCutoff, i, PtPool ) end

#Activate the following phrase when RHS figures for each subject is required.
=begin
$rgHashAr[ 1 ].each do | key, overlapRgAr |
  hh.outputPS_BW( overlapRgAr, "#{ hh.outputDirectory }/#{ key }.ps" )
  hh.outputText_BW( overlapRgAr, "#{ hh.outputDirectory }/#{ key }.txt" )
end
=end

stEdRgAr = []
for i in 0...hh.noPtPool do
  stEdRgAr << $rgHashAr[ 1 ][ ( "_%03d" % i ).to_sym ]
end
overlapRgArAr = hh.takeOverlap_C2( stEdRgAr, Mode_FoldOverlap )

#Activate the following phrase when RHS figures for fold overlaps are required.
=begin
overlapRgArAr.each_with_index do | overlapRgAr, i |
  hh.outputPS_BW( overlapRgAr, "#{ hh.outputDirectory }/HM_%03d_FoldOverlaps_PatientPool.ps" % i )
  hh.outputText_BW( overlapRgAr, "#{ hh.outputDirectory }/HM_%03d_FoldOverlaps_PatientPool.txt" % i )
end
=end

#Find out the combinations of subjects in which all pairs of subjects have overlaps at specific points on the autosomal region.
finalCount = 0; finalLevel = 2
YAML.dump( $rgHashAr[ 1 ], File.open( "#{ hh.outputDirectory }/level_1.yml", 'w' ) )
for i in 2..hh.maxLevel do
  STDERR.print Time.now.inspect + ": "
  STDOUT.print Time.now.inspect + ": "
  count = hh.hm_on_hh_PtPoolSubgroup_vs_CtPool( i, hh.noCtAllowedToShareRHS, hh.noCtAllowedToShareRCHH )
  if count == 0;
    break;
  elsif count < hh.maxNoPositiveCombinations 
    YAML.dump( $rgHashAr[ i ], File.open( "#{ hh.outputDirectory }/level_#{ i }.yml", 'w' ) )
    hh.horizontalFigure( i, "#{ hh.outputDirectory }/horizontalFigure_#{ i }.ps" )
    finalCount = count; finalLevel = i
  else
    YAML.dump( $rgHashAr[ i ], File.open( "#{ hh.outputDirectory }/level_#{ i }__terminated.yml", 'w' ) )
    hh.horizontalFigure( i, "#{ hh.outputDirectory }/horizontalFigure_#{ i }__terminated.ps" )
    finalCount = count; finalLevel = i
    break
  end
end
hh.sumSNPClear

