#include "ruby.h"
#include "math.h"
#include <HGeneC.h>
#define MODE_HM_AND_HH	1
#define	MODE_HH_ONLY		0

#if !defined (RSTRING_LEN)
#define RSTRING_LEN(x)	(RSTRING(x)->len)
#define	RSTRING_PTR(x)	(RSTRING(x)->ptr)
#define RARRAY_LEN(x)	(RARRAY(x)->len)
#define	RARRAY_PTR(x)	(RARRAY(x)->ptr)
#endif

struct summerizedSNPdata {
	int		chromosome;						//	chromosome number starts from 1.
	long	position;
	float	AFrequency;
	char	*rsID;
	double	genetMap;
	unsigned char	*gType[ 3 ];
	int	chunk;
};
struct startEnd {
	long	start;
	long	end;
};
struct OvlpSt{
	long	origin;
	int		chunk;
	int		counter;
};

static int enum_PtPool = 0, enum_CtPool = 1, enum_Extra = 2;
static int noPtPool = 0, noCtPool = 0, noExtra = 0;
static long largeGap = 0;
static long noSNP;
static struct summerizedSNPdata* sumSNP = NULL;
static struct summerizedSNPdata* saveSumSNP = NULL;
static int noChunk, noChromosome;
static struct startEnd *rgChunk, *rgChromosome;
static double aveMajorAFWOX;
//for overlap
struct OvlpSt**	ovlpStP;
struct OvlpSt*	ovlpSt;

static VALUE wrap_resetNoSNP( VALUE self, VALUE newValue ) {
	noSNP = NUM2LONG( newValue );
	return Qnil;
}

static VALUE wrap_initializeCRoutines_C( VALUE self, VALUE tNoSNP, VALUE nPtPool, VALUE nCtPool, VALUE nExtra, VALUE lgGap ){
	long i, j;
	
	// Setup global variables
	noSNP = NUM2LONG( tNoSNP );
	noPtPool = NUM2INT( nPtPool ); noCtPool = NUM2INT( nCtPool ); noExtra = NUM2INT( nExtra );
	largeGap = NUM2LONG( lgGap );
	// Allocate memory
	sumSNP = ( struct summerizedSNPdata* )xcalloc( noSNP, sizeof( struct summerizedSNPdata ) );
	for( i = 0; i < noSNP; ++i ){
		sumSNP[ i ].gType[ enum_PtPool ] = (unsigned char*)xcalloc( noPtPool + 1, sizeof( unsigned char ) );// Must be calloc to zero memory
		sumSNP[ i ].gType[ enum_CtPool ] = (unsigned char*)xcalloc( noCtPool + 1, sizeof( unsigned char ) );
		sumSNP[ i ].gType[ enum_Extra ] = (unsigned char*)xcalloc( noExtra + 1, sizeof( unsigned char ) );
		for( j = 0; j < noPtPool; ++j ){ sumSNP[ i ].gType[ enum_PtPool ][ j ] = '0'; }
		for( j = 0; j < noCtPool; ++j ){ sumSNP[ i ].gType[ enum_CtPool ][ j ] = '0'; }
		for( j = 0; j < noExtra; ++j ) { sumSNP[ i ].gType[ enum_Extra ][ j ] = '0'; }
	}
	//	overlap
	ovlpStP = ( struct OvlpSt** )xcalloc( noSNP, sizeof( struct OvlpSt* ) );
	ovlpSt = ( struct OvlpSt* )xcalloc( noSNP, sizeof( struct OvlpSt ) );
	return Qnil;
}

static VALUE wrap_readSumSNPData_C( VALUE self, VALUE dataAr, VALUE index ){
	int	chrCol = 0, posCol = 1, aFreqCol = 2, rsIDCol = 3, genetMapCol = 4, gTypePtPoolCol = 5, gTypeCtPoolCol = 6, gTypeExtraCol = 7;
	int	arrayLength, cLen;
	char *str;
	long i, j;

	i = NUM2LONG( index );
	arrayLength = (int)RARRAY_LEN( dataAr );
	
	str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + chrCol ) );
	sumSNP[ i ].chromosome = sumSNP[ i ].chunk = (int)strtol( str, NULL, 0 );
	str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + posCol ) );
	sumSNP[ i ].position = strtol( str, NULL, 0 );
	str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + aFreqCol ) );
	sumSNP[ i ].AFrequency = (float)strtod( str, NULL );
	cLen = (int)RSTRING_LEN( *( RARRAY_PTR( dataAr ) + rsIDCol ) );
	str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + rsIDCol ) );
	sumSNP[ i ].rsID = (char *)xcalloc( cLen + 1, sizeof( char ) );
	for( j = 0; j < cLen; ++j ) sumSNP[ i ].rsID[ j ] = str[ j ];
	sumSNP[ i ].rsID[ cLen ] = 0;
	str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + genetMapCol ) );
	sumSNP[ i ].genetMap = strtod( str, NULL );
	if( arrayLength >= 7 ){
		cLen = (int)RSTRING_LEN( *( RARRAY_PTR( dataAr ) + gTypePtPoolCol ) );
		str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + gTypePtPoolCol ) );
		for( j = 0; j < cLen; ++j ) sumSNP[ i ].gType[ enum_PtPool ][ j ] = str[ j ];
		sumSNP[ i ].gType[ enum_PtPool ][ cLen ] = 0;
		cLen = (int)RSTRING_LEN( *( RARRAY_PTR( dataAr ) + gTypeCtPoolCol ) );
		str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + gTypeCtPoolCol ) );
		for( j = 0; j < cLen; ++j ) sumSNP[ i ].gType[ enum_CtPool ][ j ] = str[ j ];
		sumSNP[ i ].gType[ enum_CtPool ][ cLen ] = 0;
		cLen = (int)RSTRING_LEN( *( RARRAY_PTR( dataAr ) + gTypeExtraCol ) );
		str = RSTRING_PTR( *( RARRAY_PTR( dataAr ) + gTypeExtraCol ) );
		for( j = 0; j < cLen; ++j ) sumSNP[ i ].gType[ enum_Extra ][ j ] = str[ j ];
		sumSNP[ i ].gType[ enum_Extra ][ cLen ] = 0;
	}
	return Qnil;
}

static VALUE wrap_readGenotypingData_C( VALUE self, VALUE idx, VALUE sjt, VALUE gt, VALUE pl ){
	static long aa = 0, bb = 0, ab = 0;
	volatile VALUE ar;
	long i;
	int	pool, s;
	int gT, pre;
	i = FIX2LONG( idx );
	s = FIX2INT( sjt );
	gT = FIX2INT( gt );
	pool = FIX2INT( pl );
	if ( pool == -1 ){
		ar = rb_ary_new();
		rb_ary_push( ar, LONG2FIX( aa ) );
		rb_ary_push( ar, LONG2FIX( bb ) );
		rb_ary_push( ar, LONG2FIX( ab ) );
		aa = bb = ab = 0;
		return ar;
	}
	pre = sumSNP[ i ].gType[ pool ][ s ];
	if( pre == '0' ){
		sumSNP[ i ].gType[ pool ][ s ] = gT;
		if( gT == '1' ) aa++;
		else if( gT == '2' ) bb++;
		else if( gT == '3' ) ab++;
	} else {
		if( pre == gT ) return Qnil;
		else {
			sumSNP[ i ].gType[ pool ][ s ] = '0';
			if( pre == '1' ) aa--;
			else if( pre == '2' ) bb--;
			else if( pre == '3' ) ab--;
		}
	}
	return Qnil;
}

static VALUE wrap_setupRgChunkAndRgChromosome_C( VALUE self ) {	
	long	i, j;
	int		chunk, flag, curChromosome;
	VALUE	ar, outAr, tAr;

	for( i = 0, flag = chunk = 0; i < noSNP; ++ i ){
		if( flag == 1 ){ chunk++; flag = 0; }
		if( i != noSNP - 1 ){
			if( ( sumSNP[ i ].chromosome < sumSNP[ i + 1 ].chromosome ) || ( sumSNP[ i ].position + largeGap < sumSNP[ i + 1 ].position ) ){
				flag = 1;
			}
		}
		sumSNP[ i ].chunk = chunk;
	}
	noChunk = sumSNP[ noSNP - 1 ].chunk + 1;
	rgChunk = ( struct startEnd * )( xcalloc( noChunk, sizeof( struct startEnd ) ) );
	for( i = j = 0, flag = 1; i < noSNP; ++i ){
		if( flag == 1 ){ rgChunk[ j ].start = i; flag = 0; }
		if( ( i == noSNP - 1 ) || ( sumSNP[ i ].chunk < sumSNP[ i + 1 ].chunk ) ){ rgChunk[ j++ ].end = i; flag = 1; }
	}
	
	noChromosome = sumSNP[ noSNP - 1 ].chromosome;
	rgChromosome = ( struct startEnd * )( xcalloc( noChromosome, sizeof( struct startEnd ) ) );
	for( i = j = 0, flag = 1; i < noSNP; ++i ){
		if( flag == 1 ){ rgChromosome[ j ].start = i; flag = 0; }
		if( i == noSNP - 1 || sumSNP[ i ].chromosome < sumSNP[ i + 1 ].chromosome ){ rgChromosome[ j++ ].end = i; flag = 1; }
	}
	outAr = rb_ary_new();
	ar = rb_ary_new();	
	for( i = 0; i < noChunk; ++i ){
		tAr = rb_ary_new();
		rb_ary_push( tAr, LONG2FIX( rgChunk[ i ].start ) );
		rb_ary_push( tAr, LONG2FIX( rgChunk[ i ].end ) );
		rb_ary_push( ar, tAr );
	}
	rb_ary_push( outAr, ar );
	ar = rb_ary_new();
	for( i = 0; i < noChromosome; ++i ){
		tAr = rb_ary_new();
		rb_ary_push( tAr, LONG2FIX( rgChromosome[ i ].start ) );
		rb_ary_push( tAr, LONG2FIX( rgChromosome[ i ].end ) );
		rb_ary_push( ar, tAr );
	}
	rb_ary_push( outAr, ar );
	return outAr;
}

static VALUE wrap_getSumSNPMem_C( VALUE self, VALUE idx ){	
	long	i;
	char	buf[ 2000 ];
	volatile VALUE ar;
	
	i = FIX2LONG( idx );
	
	ar = rb_ary_new();
	sprintf( buf, "%d", sumSNP[ i ].chromosome );
	rb_ary_push( ar, rb_str_new2( buf ) );
	sprintf( buf, "%ld", sumSNP[ i ].position );
	rb_ary_push( ar, rb_str_new2( buf ) );
	sprintf( buf, "%f", sumSNP[ i ].AFrequency );
	rb_ary_push( ar, rb_str_new2( buf ) );
	rb_ary_push( ar, rb_str_new2( sumSNP[ i ].rsID ) );
	sprintf( buf, "%f", sumSNP[ i ].genetMap );
	rb_ary_push( ar, rb_str_new2( buf ) );
	rb_ary_push( ar, rb_str_new2( sumSNP[ i ].gType[ 0 ] ) );
	rb_ary_push( ar, rb_str_new2( sumSNP[ i ].gType[ 1 ] ) );
	rb_ary_push( ar, rb_str_new2( sumSNP[ i ].gType[ 2 ] ) );
	sprintf( buf, "%d", sumSNP[ i ].chunk );
	rb_ary_push( ar, rb_str_new2( buf ) );
	return ar;
}

static VALUE wrap_correctErrorHM_C( VALUE self, VALUE eRate ) {
	int	noRepeat = 4;
	
	struct	heteroSNP {
		long	index;
		double	score;
	} *hetSNP;
	
	int	iChromosome, iPool, iSt, j, k;
	double errorRate, factor;
	long noSNPRemoved, noHetSNP, *bucket, scoreThreshold;
	int noSt[ 3 ];
	long	i;
	
	noSt[ 0 ] = noPtPool; noSt[ 1 ] = noCtPool, noSt[ 2 ] = noExtra;
	fprintf( stderr, "Correcting genotype data...\n" );
	errorRate = rb_num2dbl( eRate );
	for( iPool = enum_PtPool; iPool <= enum_Extra; ++iPool ){
		for( iSt = 0; iSt < noSt[ iPool ]; ++ iSt ){
			for( iChromosome = 0; iChromosome < noChromosome; ++iChromosome ){
				noSNPRemoved = errorRate * ( rgChromosome[ iChromosome ].end - rgChromosome[ iChromosome ].start + 1 ) * ( aveMajorAFWOX * aveMajorAFWOX + ( 1.0 - aveMajorAFWOX ) * ( 1.0 - aveMajorAFWOX ) ); 
				for( k = 0; k < noRepeat; ++k ){
					for( noHetSNP = 0, i = rgChromosome[ iChromosome ].start; i <= rgChromosome[ iChromosome ].end; ++i ){;
						if( sumSNP[ i ].gType[ iPool ][ iSt ] == '3' ){
							noHetSNP++;
						}
					}
					if( noHetSNP < noSNPRemoved / noRepeat ) break;	// For chromosome X in males.
					hetSNP = (struct heteroSNP*)xcalloc( noHetSNP, sizeof( struct heteroSNP ) );
					for( j = 0, i = rgChromosome[ iChromosome ].start; i <= rgChromosome[ iChromosome ].end; ++i ){
						if( sumSNP[ i ].gType[ iPool ][ iSt ] == '3' ) hetSNP[ j++ ].index = i;
					}
					for( i = 0; i < noHetSNP; ++i ){
						hetSNP[ i ].score = sumSNP[ hetSNP[ ( i < noHetSNP - 1 )? i + 1: i ].index ].genetMap - sumSNP[ hetSNP[ ( i > 0 )? i - 1: i ].index ].genetMap;
					}
					qsort( hetSNP, noHetSNP, sizeof( struct heteroSNP ), compareHeteroSNP );
					for( i = 0; i < noSNPRemoved / noRepeat; ++i ){
						sumSNP[ hetSNP[ i ].index ].gType[ iPool ][ iSt ] = '0';
					}
					free( hetSNP );
				}
			}
		}
	}
	return rb_float_new( errorRate );
}

static int compareHeteroSNP( const void *a, const void *b ){
	struct	heteroSNP {
		long	index;
		long	score;
	};
	int ret;
	if( ( (struct heteroSNP*)a )->score < ( (struct heteroSNP*)b )->score ) ret = 1;
	else if( ( (struct heteroSNP*)a )->score > ( (struct heteroSNP*)b )->score ) ret = -1;
	else ret = 0;
	return ret;
}

static int compare_ovlpSt( const void *a, const void *b ){
	//ascending order
	if( ( (struct OvlpSt*)a )->origin < ( (struct OvlpSt*)b )->origin ) return -1;
	else if( ( (struct OvlpSt*)a )->origin > ( (struct OvlpSt*)b )->origin ) return 1;
	else return 0;
}

static VALUE wrap_obtainRHS_C( VALUE self, VALUE ctOff, VALUE st, VALUE pl ){
	int				j, subject, pool;
	volatile VALUE	ar, sted;
	long			i, start, end;
	double			cutOff;
	
	cutOff = rb_num2dbl( ctOff ); subject = FIX2INT( st ); pool = FIX2INT( pl );
	ar = rb_ary_new();
	for( j = 0; j < noChunk; ++j ){
		for( start = rgChunk[ j ].start, i = start + 1; i <= rgChunk[ j ].end; ++i ){
			if( sumSNP[ i ].gType[ pool ][ subject ] == '3' || i == rgChunk[ j ].end ){
				end = i;
				if( sumSNP[ end ].genetMap - sumSNP[ start ].genetMap >= cutOff ){
					sted = rb_ary_new();
					rb_ary_push( sted, LONG2FIX( start ) );
					rb_ary_push( sted, LONG2FIX( end ) );
					rb_ary_push( ar, sted );
				}
				start = end;
			}
		}
	}
	return ar;
}

static VALUE wrap_sumSNPClear( VALUE self ){
	long i;
	for( i = 0; i < noSNP; ++i ){ free( sumSNP[ i ].gType[ 0 ] ); free( sumSNP[ i ].gType[ 1 ] ); free( sumSNP[ i ].gType[ 2 ] ); }
	if ( sumSNP != NULL ){
		free( sumSNP );
	}
	return Qnil;
}

static VALUE wrap_isHMandHHSameInSelectedPt( VALUE self, VALUE idx, VALUE stListAr, VALUE md ){
	// if mode == 1 HM and HH.  if mode == 0, HH only
	int	noSt, i, aa, bb, ab, mode;
	long	index;
	int *stNo;
	char *ptGType;

	noSt = (int)RARRAY_LEN( stListAr );
	index = NUM2LONG( idx ); mode = NUM2INT( md );
	if( noSt == 0 ){ fprintf( stderr, "Patient list is empty: isHMandHHSame.\n" ); return Qfalse; }
	stNo = (int *)xcalloc( noSt, sizeof( int ) );
	ptGType = (char *)xmalloc( noSt );
	for( i = 0; i < noSt; ++i ){
		stNo[ i ] = NUM2INT( *( RARRAY_PTR( stListAr ) + i ) );
		ptGType[ i ] = sumSNP[ index ].gType[ enum_PtPool ][ stNo[ i ] ];
	}
	for( aa = bb = ab = i = 0; i < noSt; ++i ){
		switch( ptGType[ i ] ) {
			case '1': aa++; break;
			case '2': bb++; break;
			case '3': ab++; break;
		}
	}
	if( mode == MODE_HM_AND_HH ){ // HM incompatible
		if( aa != 0 && bb != 0 ) { free( stNo ); free( ptGType ); return Qfalse; }
		if( ab != 0 ) { free( stNo ); free( ptGType ); return Qfalse; }
	} else {
		if( aa != 0 && bb != 0 ) { free( stNo ); free( ptGType ); return Qfalse; }
	}	//	HH incompatible
	free( stNo ); free( ptGType ); return Qtrue;
}

static VALUE wrap_countOfRCHHInCtPool( VALUE self, VALUE range, VALUE stListAr, VALUE cOff ){
	long	i, j, noSt, rchhRgStart, rchhRgEnd, aa, bb;
	int 	*stNo, count, flag, cutoff;
	char 	*ptGType, *ancestralHH, cAns, cCt, buf[ 100 ];
	
	noSt = (int)RARRAY_LEN( stListAr );
	stNo = (int *)xcalloc( noSt, sizeof( int ) );
	cutoff = NUM2INT( cOff );
	ptGType = (char *)xmalloc( noSt );
	for( i = 0; i < noSt; ++i ){
		stNo[ i ] = NUM2INT( *( RARRAY_PTR( stListAr ) + i ) );
	}
	rchhRgStart = NUM2INT( *RARRAY_PTR( range ) );
	rchhRgEnd = NUM2INT( *( RARRAY_PTR( range ) + 1 ) );
	fprintf(stderr, "%ld %ld\n",rchhRgStart, rchhRgEnd );
	ancestralHH = (char *)xmalloc( rchhRgEnd - rchhRgStart + 2 );
	for( j = 1; j < rchhRgEnd - rchhRgStart; ++j ){
		for( i = 0; i < noSt; ++i ){
			ptGType[ i ] = sumSNP[ j + rchhRgStart ].gType[ enum_PtPool ][ stNo[ i ] ];
		}
		for( aa = bb = i = 0; i < noSt; ++i ){
			switch( ptGType[ i ] ) {
				case '1': aa++; break;
				case '2': bb++; break;
			}
		}
		if( aa != 0 ) ancestralHH[ j ] = '1';
		else if ( bb != 0 ) ancestralHH[ j ] = '2';
		else ancestralHH[ j ] = '0';
	}
	ancestralHH[ j ] = 0;
	for( i = 0, count = 0; i < noCtPool; ++i ){
		for( flag = 0, j = 1; j < rchhRgEnd - rchhRgStart; ++j ){	// Boundaries are not included
			cAns = ancestralHH[ j ]; cCt = sumSNP[ j + rchhRgStart ].gType[ enum_CtPool ][ i ];
			if ( cAns == '1' && cCt == '2' || cAns == '2' && cCt == '1' ){
				flag = 1; break;
			}
		}
		if( flag == 0 ) count++;
		if( count > cutoff ) {
			free( ptGType ); free( ancestralHH ); return Qfalse;
		}
	}
	free( ptGType ); free( ancestralHH );
	return Qtrue;
}

static VALUE wrap_takeOverlap_C2( VALUE self, VALUE stEdRgArAr, VALUE eqToOrMoreThan ){
	volatile	VALUE	stEdRgAr, stEdRg, outerAr, innerAr, outmostAr;
	long	i, j, k, st, ed, preEnd, index;
	int	*ar, equalToOrMoreThan, stFoldOverlap, edFoldOverlap, in, preVal, no_stEdRgArAr, no_stEdRgAr;
	struct OvlpSt *stP, *edP, *iP, *jP;
	
	no_stEdRgArAr = (int)RARRAY_LEN( stEdRgArAr ); 		// no_stEdRgArAr == 1 when [[]].
	equalToOrMoreThan = NUM2INT( eqToOrMoreThan );
	for( index = i = 0; i < no_stEdRgArAr; ++i ){
		if( Qnil == ( stEdRgAr = *( RARRAY_PTR( stEdRgArAr ) + i ) ) ) { return rb_ary_new(); };
		no_stEdRgAr = (int)RARRAY_LEN( stEdRgAr );		// no_stEdRgAr == 0 when stEdRgArAr == [[]]
		if( no_stEdRgAr == 0 ){ return rb_ary_new(); }
			for( preEnd = -1, j = 0; j < no_stEdRgAr; ++j ){
			stEdRg = *( RARRAY_PTR( stEdRgAr ) + j );
			st = NUM2LONG( *RARRAY_PTR( stEdRg ) );
			// Add the point before a start of region, and the point at the start of a region
			if( st != 0 && sumSNP[ st ].chunk == sumSNP[ st - 1 ].chunk ){
				ovlpStP[ st - 1 ] = ovlpSt + index; index++;
				ovlpStP[ st - 1 ]->origin = st - 1;
				ovlpStP[ st - 1 ]->chunk = sumSNP[ st ].chunk;
			}
			ovlpStP[ st ] = ovlpSt + index; index++;
			ovlpStP[ st ]->origin = st;
			ovlpStP[ st ]->chunk = sumSNP[ st ].chunk;
			ed = NUM2LONG( *( RARRAY_PTR( stEdRg ) + 1 ) );
			// Add the point after a end of region, and the point at the end of a region
			if( ed != noSNP - 1 && sumSNP[ ed ].chunk == sumSNP[ ed + 1 ].chunk ){
				ovlpStP[ ed + 1 ] = ovlpSt + index; index++;
				ovlpStP[ ed + 1 ]->origin = ed + 1;
				ovlpStP[ ed + 1 ]->chunk = sumSNP[ ed ].chunk;
			}
			ovlpStP[ ed ] = ovlpSt + index; index++;
			ovlpStP[ ed ]->origin = ed;
			ovlpStP[ ed ]->chunk = sumSNP[ ed ].chunk;
		}
	}
	// Add the boundaries of a chunk.
	for( i = 0; i < noChunk; ++i ){
		ovlpStP[ rgChunk[ i ].start ] = ovlpSt + index; index++;
		ovlpStP[ rgChunk[ i ].start ]->origin = rgChunk[ i ].start;
		ovlpStP[ rgChunk[ i ].start ]->chunk = sumSNP[ rgChunk[ i ].start ].chunk;
		ovlpStP[ rgChunk[ i ].end ] = ovlpSt + index; index++;
		ovlpStP[ rgChunk[ i ].end ]->origin = rgChunk[ i ].end;
		ovlpStP[ rgChunk[ i ].end ]->chunk = sumSNP[ rgChunk[ i ].end ].chunk;
	}
	// sort intOvlpSt
	qsort( ovlpSt, index, sizeof( struct OvlpSt ), compare_ovlpSt );
	// remove duplicate
	for( j = i = 0; i < index - 1; ++ i ){
		//fprintf( stderr, "%ld ", ovlpSt[ i ].origin );
		if( ovlpSt[ i ].origin == ovlpSt[ i + 1 ].origin ){
			ovlpSt[ i ].origin = 10000000; j++;
		}
	}
	qsort( ovlpSt, index, sizeof( struct OvlpSt ), compare_ovlpSt );
	index -= j;
	// recover pointer
	for( i = 0; i < index; ++i ){
		ovlpStP[ ovlpSt[ i ].origin ] = ovlpSt + i;
		ovlpSt[ i ].counter = 0;
	}
	// count
	for( i = 0; i < no_stEdRgArAr; ++i ){
		stEdRgAr = *( RARRAY_PTR( stEdRgArAr ) + i );
		no_stEdRgAr = (int)RARRAY_LEN( stEdRgAr );
		for( j = 0; j < no_stEdRgAr; ++j ){
			stEdRg = *( RARRAY_PTR( stEdRgAr ) + j );
			stP = ovlpStP[ NUM2LONG( *RARRAY_PTR( stEdRg ) ) ];
			edP = ovlpStP[ NUM2LONG( *( RARRAY_PTR( stEdRg ) + 1 ) ) ];
			for( iP = stP; iP <= edP; ++iP ){
				iP->counter++;
			}
		}
	}
	// multiple color mode when equalToOrMoreThan == -2
	if( equalToOrMoreThan == -2 ){
		outerAr = rb_ary_new();
		for( i = 0; i < noChunk; ++i ){
			for( stP = jP = ovlpStP[ rgChunk[ i ].start ], preVal = jP->counter; jP <= ovlpStP[ rgChunk[ i ].end ]; ++jP ){
				if( preVal == jP->counter ){
					if( jP != ovlpStP[ rgChunk[ i ].end ] ){ 
						continue;
					} else {
						edP = jP;
						innerAr = rb_ary_new();
						// return [ [ st, value ], [ ed, value ] ]
						rb_ary_push( innerAr, LONG2FIX( stP->origin ) ); rb_ary_push( innerAr, LONG2FIX( preVal ) );
						rb_ary_push( outerAr, innerAr );
						innerAr = rb_ary_new();
						rb_ary_push( innerAr, LONG2FIX( edP->origin ) ); rb_ary_push( innerAr, LONG2FIX( preVal ) );
						rb_ary_push( outerAr, innerAr );
					}
				} else {
					edP = jP - 1;
					innerAr = rb_ary_new();
					rb_ary_push( innerAr, LONG2FIX( stP->origin ) ); rb_ary_push( innerAr, LONG2FIX( preVal ) );
					rb_ary_push( outerAr, innerAr );
					innerAr = rb_ary_new();
					rb_ary_push( innerAr, LONG2FIX( edP->origin ) ); rb_ary_push( innerAr, LONG2FIX( preVal ) );
					rb_ary_push( outerAr, innerAr );
					stP = jP; preVal = jP->counter;
				}
			}
		}
		return outerAr;		
	}
	// Fold overlap mode
	if( equalToOrMoreThan == -1 ){
		outmostAr = rb_ary_new(); rb_ary_push( outmostAr, rb_ary_new() );
		stFoldOverlap = 1; edFoldOverlap = no_stEdRgArAr;
	} else {
		stFoldOverlap = edFoldOverlap = equalToOrMoreThan;
	}
	for( k = stFoldOverlap; k <= edFoldOverlap; ++k ){
		// return [ [], [ [ st, ed ], [ st, ed ], [ st, ed ]... ], [ [ st, ed ], [ st, ed ], [ st, ed ]... ] ]
		outerAr = rb_ary_new();
		for( i = 0; i < noChunk; ++i ){
			for( in = 0, jP = ovlpStP[ rgChunk[ i ].start ]; jP <= ovlpStP[ rgChunk[ i ].end ]; ++jP ){
				if( in == 0 ){
					if( jP->counter < k ){ continue; }
					else {
						if( jP != ovlpStP[ rgChunk[ i ].end ] ){ 
							stP = jP; in = 1;
						} else {
							stP = edP = jP;
							innerAr = rb_ary_new();
							rb_ary_push( innerAr, LONG2FIX( stP->origin ) ); rb_ary_push( innerAr, LONG2FIX( edP->origin ) );
							rb_ary_push( outerAr, innerAr );
						}
					}
				} else {
					if( jP->counter >= k ){
						if( jP != ovlpStP[ rgChunk[ i ].end ] ) continue;
						edP = jP;
					} else {
						edP = jP - 1; in = 0;
					}
					innerAr = rb_ary_new();
					rb_ary_push( innerAr, LONG2FIX( stP->origin ) ); rb_ary_push( innerAr, LONG2FIX( edP->origin ) );
					rb_ary_push( outerAr, innerAr );
				}
			}
		}
		// Fold overlap mode
		if( equalToOrMoreThan == -1 ){
			rb_ary_push( outmostAr, outerAr );
		}	
	}
	// Fold overlap mode
	if( equalToOrMoreThan == -1 ){
		return outmostAr;
	} else {
		return outerAr;		
	}
}

static VALUE wrap_significantQuadriadRegion_C( VALUE self, VALUE stEdRgArAr, VALUE stListAr, VALUE cOffHM, VALUE cOffHH, VALUE omitComparingRHS, VALUE omitComparingRCHH ){
	volatile VALUE ar, stEdRgAr, tAr;
	int	no_stEdRgArAr, flag, noSt, *stNo, cutoffHM, cutoffHH;
	long	st, ed, rgStart, rgEnd, i, rSt, rEd;

	cutoffHM = NUM2INT( cOffHM ); cutoffHH = NUM2INT( cOffHH );
	noSt = (int)RARRAY_LEN( stListAr );
	stNo = (int *)xcalloc( noSt, sizeof( int ) );
	for( i = 0; i < noSt; ++i ){
		stNo[ i ] = NUM2INT( *( RARRAY_PTR( stListAr ) + i ) );
	}
	ar = rb_ary_new();
	no_stEdRgArAr = (int)RARRAY_LEN( stEdRgArAr );
	for( i = 0; i < no_stEdRgArAr; ++i ){ // Select RHS with the same HH, then check its significance
		if( Qnil == ( stEdRgAr = *( RARRAY_PTR( stEdRgArAr ) + i ) ) ) { free( stNo ); return rb_ary_new(); }
		// Fail safe.  If one of the stEdRgAr is nil, then return []
		rgStart = NUM2LONG( *RARRAY_PTR( stEdRgAr ) ); rgEnd = NUM2LONG( *( RARRAY_PTR( stEdRgAr ) + 1 ) );
		ed = rgStart;
		flag = 0; st = ed;
		while( 1 ){
			 if( isBothInRHSAndRCHH( ed, noSt, stNo ) == 0 ){ // mode = HMAndHH
				if( flag == 1 ){
					if( isSignificantQuadriad( st, ed, &rSt, &rEd, noSt, stNo, cutoffHM, cutoffHH, omitComparingRHS, omitComparingRCHH ) == 1 ){
						tAr = rb_ary_new();
						rb_ary_push( tAr, LONG2FIX( st ) ); rb_ary_push( tAr, LONG2FIX( ed ) );
						rb_ary_push( tAr, LONG2FIX( rSt ) ); rb_ary_push( tAr, LONG2FIX( rEd ) );
						rb_ary_push( ar, tAr );
					}
				}
				if( ed == rgEnd ){ break; }
				else {
					if( st == ed ){ ed++; }
					flag = 0; st = ed;
				}
			}else{
				if( flag == 0 && rgStart < st ){ st--; }
				flag = 1;
				if( ed == rgEnd ){
					if( flag == 1 ){
						if( isSignificantQuadriad( st, ed, &rSt, &rEd, noSt, stNo, cutoffHM, cutoffHH, omitComparingRHS, omitComparingRCHH ) == 1 ){
							tAr = rb_ary_new();
							rb_ary_push( tAr, LONG2FIX( st ) ); rb_ary_push( tAr, LONG2FIX( ed ) );
							rb_ary_push( tAr, LONG2FIX( rSt ) ); rb_ary_push( tAr, LONG2FIX( rEd ) );
							rb_ary_push( ar, tAr );
						}
					}
					break;
				} else { ed++; }
			}
		}
	}
	free( stNo );
	return ar;
}

static int isSignificantQuadriad( long rgStart, long rgEnd, long *rSt, long *rEd, int noSt, int *stNo, int cutoffHM, int cutoffHH, VALUE omitComparingRHS, VALUE omitComparingRCHH ){
	long stPt, i, j, rchhRgStart, rchhRgEnd, rhsRgStart, rhsRgEnd, st, ed;
	char *ancestralHH, *ancestralH, *ptGType, cAns, cCt;
	int	flag, count, chunkNo, aa, bb;
	
	ptGType = (char *)xmalloc( noSt );
	rhsRgStart = rgStart + 1; rhsRgEnd = rgEnd - 1;
//Define ancestral haplotype in RHS
	ancestralH = (char *)xmalloc( rhsRgEnd - rhsRgStart + 2 );
	for( j = 1; j < rhsRgEnd - rhsRgStart; ++j ){
		for( i = 0; i < noSt; ++i ){
			ptGType[ i ] = sumSNP[ j + rhsRgStart ].gType[ enum_PtPool ][ stNo[ i ] ];
		}
		for( i = 0, aa = bb = 0; i < noSt; ++i ){
			switch( ptGType[ i ] ) {
				case '1': aa++; break;
				case '2': bb++; break;
			}
		}
		if( aa != 0 ) ancestralH[ j ] = '1';
		else if ( bb != 0 ) ancestralH[ j ] = '2';
		else ancestralH[ j ] = '0';
	}
	ancestralH[ j ] = 0;
	//At this point, ancestoral haplotype in the RHS is determined.
	//Define ancestral homozygosity haplotype in RCHH
	chunkNo = sumSNP[ rgStart ].chunk;
	stPt = rgStart;
	while( isBothInRHSAndRCHH( stPt, noSt, stNo ) == 0 ) { stPt++; }	// Find out a place to start
	st = stPt;
	while ( isInRCHH( st, noSt, stNo ) == 1 && rgChunk[ chunkNo ].start < st ) { st--; } // Find out left end
	ed = stPt;
	while ( isInRCHH( ed, noSt, stNo ) == 1 && ed < rgChunk[ chunkNo ].end ) { ed++; } // Find out right end
	*rSt = rchhRgStart = st; *rEd = rchhRgEnd = ed;
	ancestralHH = (char *)xmalloc( rchhRgEnd - rchhRgStart + 2 );
	for( j = 1; j < rchhRgEnd - rchhRgStart; ++j ){	// Boundaries are excluded.
		for( i = 0; i < noSt; ++i ){
			ptGType[ i ] = sumSNP[ j + rchhRgStart ].gType[ enum_PtPool ][ stNo[ i ] ];
		}
		for( i = 0, aa = bb = 0; i < noSt; ++i ){
			switch( ptGType[ i ] ) {
				case '1': aa++; break;
				case '2': bb++; break;
			}
		}
		if( aa != 0 ) ancestralHH[ j ] = '1';
		else if ( bb != 0 ) ancestralHH[ j ] = '2';
		else ancestralHH[ j ] = '0';
	}
	ancestralHH[ j ] = 0;
	//At this point, ancestoral haplotype in the RCHH is determined.
	if( omitComparingRCHH == Qfalse ){
		for( i = 0, count = 0; i < noCtPool; ++i ){	// Check RCHH.
			for( flag = 0, j = 1; j < rchhRgEnd - rchhRgStart; ++j ){	// Boundaries are Boundaries are excluded.
				cAns = ancestralHH[ j ]; cCt = sumSNP[ j + rchhRgStart ].gType[ enum_CtPool ][ i ];
				if ( cAns == '1' && cCt == '2' || cAns == '2' && cCt == '1' ){
					flag = 1; break;
				}
			}
			if( flag == 0 ) count++;
			if( count > cutoffHH ) {
				free( ptGType ); free( ancestralHH ); free( ancestralH ); return 0;  // Discarded.
			}
		}
	}
	if( omitComparingRHS == Qfalse ){
		for( i = 0, count = 0; i < noCtPool; ++i ){	// Check RHS.
			for( flag = 0, j = 1; j < rhsRgEnd - rhsRgStart; ++j ){	// Boundaries are Excluded
				cAns = ancestralH[ j ]; cCt = sumSNP[ j + rhsRgStart ].gType[ enum_CtPool ][ i ];
				if ( cCt == '3' || cAns == '1' && cCt == '2' || cAns == '2' && cCt == '1' ){  
					flag = 1; break;
				}
			}
			if( flag == 0 ) count++;
			if( count > cutoffHM ) {
				free( ptGType ); free( ancestralHH ); free( ancestralH ); return 0;  // Discarded.
			}
		}
	}
	free( ptGType ); free( ancestralHH ); free( ancestralH );
	return 1;  // Top cutoffHH/noCtPool && top cutoffHM/noCtPool.
}

static int isBothInRHSAndRCHH( long index, int noSt, int* stNo ){
// index...SNP no in sumSNP struct.  noSt...number of subject to study.  stNo...list of the subject. Number starts from zero.
	int	i, aa, bb, ab;
	// aa..number of subjects with AA genotype.  bb..with BB genotype.  ab..with AB genotype.
	if( noSt == 0 ){ fprintf( stderr, "Patient list is empty: isHMandHHSame.\n" ); return 0; }
	for( aa = bb = ab = i = 0; i < noSt; ++i ){
		switch( sumSNP[ index ].gType[ enum_PtPool ][ stNo[ i ] ] ){
			case '1': aa++; break;
			case '2': bb++; break;
			case '3': ab++; break;
		}
	}
	if( ( aa != 0 && bb != 0 ) || ( ab != 0 ) ) return 0; else return 1;
	//	If genotype of one subject is AA and one subject is BB, or some subjects have AB, then HH in RHS is not conserved.
	//	Return false = 0 if not conserved.  Return true = 1 if conserved.
}

static int isInRCHH( long index, int noSt, int* stNo ){
// index...SNP no in sumSNP struct.  noSt...number of subject to study.  stNo...list of the subject. Number starts from zero.
	int	i, aa, bb, ab;
	// aa..number of subjects with AA genotype.  bb..with BB genotype.  ab..with AB genotype.
	if( noSt == 0 ){ fprintf( stderr, "Patient list is empty: isHMandHHSame.\n" ); return 0; }
	for( aa = bb = ab = i = 0; i < noSt; ++i ){
		switch( sumSNP[ index ].gType[ enum_PtPool ][ stNo[ i ] ] ){
			case '1': aa++; break;
			case '2': bb++; break;
			case '3': ab++; break;
		}
	}
	if( aa != 0 && bb != 0 ) return 0; else return 1;
	//	If genotype of one subject is AA and one subject is BB, return false = 0
	//	If not, return true = 1
}

static VALUE wrap_insertGenotype( VALUE self, VALUE generationM, VALUE snpNoCausingDisease, VALUE noPtInsertedFragmentIBD, VALUE backgroundPercentageOfIBD ){
	int	genM, nPtIns, bkgPercentage, i, count, *fPt, *stNo, r, gTMRCA, gTSt, chunkNo, cur;
	long snpNoDis, lowEnd1, lowEnd2, highEnd1, highEnd2, j, temp;
	double	lambda, randFrag, bkgRate;
	
	genM = NUM2INT( generationM );
	snpNoDis = NUM2LONG( snpNoCausingDisease );
	nPtIns = NUM2INT( noPtInsertedFragmentIBD );
	bkgPercentage = NUM2INT( backgroundPercentageOfIBD );
	init_genrand( time( 0 ) );

	insertFragmentIBD_Hetero( enum_PtPool, noPtPool, (int)( noPtPool * bkgPercentage / 100.0 ), genM, snpNoDis );
	insertFragmentIBD_Hetero( enum_CtPool, noCtPool, (int)( noCtPool * bkgPercentage / 100.0 ), genM, snpNoDis );
	insertFragmentIBD_Homo( enum_PtPool, noPtPool, nPtIns, genM, snpNoDis );
	insertFragmentIBD_Homo( enum_CtPool, noCtPool, (int)( noCtPool * bkgPercentage * bkgPercentage / 10000.0 ), genM, snpNoDis );
	return Qnil;
}

static void insertFragmentIBD_Homo( int pool, int noStInPool, int noStToInsert, int genM, long snpNoDis ){
	double lambda, randFrag;
	long	cur, lowEnd1, highEnd1, lowEnd2, highEnd2, i, j, temp;
	int		gTMRCA, gTSt, *fPt, *stNo, r, chunkNo;
	
	fPt = (int *)xcalloc( noStInPool, sizeof( int ) );				//Flag
	stNo = (int *)xcalloc( noStToInsert, sizeof( int ) );			//For recording subject number to insert fragment.
	for( i = 0; i < noStInPool; ++i ){ fPt[ i ] = 0; }				//Zero flags.
	for( i = 0; i < noStToInsert; ){
		r = (int)( genrand_real1() * noStInPool* 3 - noStInPool );			// -noStInPool < r < 2 * noStInPool
		if( r >= 0 && r < noStInPool ) {
			if( fPt[ r ] == 0 ){ fPt[ r ] = 1; stNo[ i++ ] = r; }	// Record selected Pts.
		}
	}
	for( i = 0; i < noStToInsert; ++i ){
		lambda = genM / 100.0;										// 1/lambda is equal to the average length of fragment.
		chunkNo = sumSNP[ snpNoDis ].chunk;							// Current chunk
		randFrag = -log( 1.0 - genrand_real3() ) / lambda;			// randFrag provides random number conforming to an exponential distribution.
		for( cur = snpNoDis; ( cur != rgChunk[ chunkNo ].start ) && ( sumSNP[ snpNoDis ].genetMap - sumSNP[ cur ].genetMap <= randFrag ); cur-- ) {}
		if( sumSNP[ snpNoDis ].genetMap - sumSNP[ cur ].genetMap > randFrag ) { lowEnd1 = cur + 1; } else { lowEnd1 = cur; }
		randFrag = -log( 1.0 - genrand_real3() ) / lambda;			// randFrag provides random number conforming to an exponential distribution.
		for( cur = snpNoDis; ( cur != rgChunk[ chunkNo ].start ) && ( sumSNP[ snpNoDis ].genetMap - sumSNP[ cur ].genetMap <= randFrag ); cur-- ) {}
		if( sumSNP[ snpNoDis ].genetMap - sumSNP[ cur ].genetMap > randFrag ) { lowEnd2 = cur + 1; } else { lowEnd2 = cur; }

		randFrag = -log( 1.0 - genrand_real3() ) / lambda;
		for( cur = snpNoDis; ( cur != rgChunk[ chunkNo ].end ) && ( sumSNP[ cur ].genetMap - sumSNP[ snpNoDis ].genetMap <= randFrag ); cur++ ) {}
		if( sumSNP[ cur ].genetMap - sumSNP[ snpNoDis ].genetMap > randFrag ) { highEnd1 = cur - 1; } else { highEnd1 = cur; }
		randFrag = -log( 1.0 - genrand_real3() ) / lambda;
		for( cur = snpNoDis; ( cur != rgChunk[ chunkNo ].end ) && ( sumSNP[ cur ].genetMap - sumSNP[ snpNoDis ].genetMap <= randFrag ); cur++ ) {}
		if( sumSNP[ cur ].genetMap - sumSNP[ snpNoDis ].genetMap > randFrag ) { highEnd2 = cur - 1; } else { highEnd2 = cur; }

		if( lowEnd1 > lowEnd2 ){ temp = lowEnd2; lowEnd2 = lowEnd1; lowEnd1 = temp; }
		if( highEnd1 > highEnd2 ){ temp = highEnd2; highEnd2 = highEnd1; highEnd1 = temp; }

		for( j = lowEnd1; j < lowEnd2; ++ j ){
			gTMRCA = sumSNP[ j ].gType[ 2 ][ 0 ];
			gTSt = sumSNP[ j ].gType[ pool ][ stNo[ i ] ];
			if( ( gTMRCA == '1' && gTSt == '2' ) || ( gTMRCA == '2' && gTSt == '1' ) ) { sumSNP[ j ].gType[ pool ][ stNo[ i ] ] = gTMRCA; }			
		}
		for( j = lowEnd2; j <= highEnd1; ++ j ){
			gTMRCA = sumSNP[ j ].gType[ 2 ][ 0 ];
			gTSt = sumSNP[ j ].gType[ pool ][ stNo[ i ] ];
			if( gTMRCA == '1' || gTMRCA == '2' ){ sumSNP[ j ].gType[ pool ][ stNo[ i ] ] = gTMRCA; }
			else if( gTMRCA == '3' ){ sumSNP[ j ].gType[ pool ][ stNo[ i ] ] = '1'; }
			else { sumSNP[ j ].gType[ pool ][ stNo[ i ] ] = '0'; }
		}
		for( j = highEnd1 + 1; j <= highEnd2; ++ j ){
			gTMRCA = sumSNP[ j ].gType[ 2 ][ 0 ];
			gTSt = sumSNP[ j ].gType[ pool ][ stNo[ i ] ];
			if( ( gTMRCA == '1' && gTSt == '2' ) || ( gTMRCA == '2' && gTSt == '1' ) ) { sumSNP[ j ].gType[ pool ][ stNo[ i ] ] = gTMRCA; }			
		}
		if ( pool == 0 ) {
		//	fprintf( stderr, "Inserted in both strands PtPool %d: %ld, %ld, %ld %ld\n", stNo[ i ], lowEnd1, lowEnd2, highEnd1, highEnd2 );
			fprintf( stdout, "Inserted in both strands PtPool %d: %ld, %ld, %ld %ld\n", stNo[ i ], lowEnd1, lowEnd2, highEnd1, highEnd2 );
		} else {
		//	fprintf( stderr, "Inserted in both strands CtPool %d: %ld, %ld, %ld %ld\n", stNo[ i ], lowEnd1, lowEnd2, highEnd1, highEnd2 );
			fprintf( stdout, "Inserted in both strands CtPool %d: %ld, %ld, %ld %ld\n", stNo[ i ], lowEnd1, lowEnd2, highEnd1, highEnd2 );
		}
	}
	free( fPt ); free( stNo );
}

static void insertFragmentIBD_Hetero( int pool, int noStInPool, int noStToInsert, int genM, long snpNoDis ){
	double lambda, randFrag;
	long	cur, lowEnd1, highEnd1, i, j;
	int		gTMRCA, gTSt, *fPt, *stNo, r, chunkNo;
	
	fPt = (int *)xcalloc( noStInPool, sizeof( int ) );				//Flag
	stNo = (int *)xcalloc( noStToInsert, sizeof( int ) );			//For recording subject number to insert fragment.
	for( i = 0; i < noStInPool; ++i ){ fPt[ i ] = 0; }				//Zero flags.
	for( i = 0; i < noStToInsert; ){
		r = (int)( genrand_real1() * noStInPool* 3 - noStInPool );			// -noStInPool < r < 2 * noStInPool
		if( r >= 0 && r < noStInPool ) {
			if( fPt[ r ] == 0 ){ fPt[ r ] = 1; stNo[ i++ ] = r; }	// Record selected Pts.
		}
	}
	for( i = 0; i < noStToInsert; ++i ){
		lambda = genM / 100.0;										// 1/lambda is equal to the average length of fragment.
		chunkNo = sumSNP[ snpNoDis ].chunk;							// Current chunk
		randFrag = -log( 1.0 - genrand_real3() ) / lambda;			// randFrag provides random number conforming to an exponential distribution.
		for( cur = snpNoDis; ( cur != rgChunk[ chunkNo ].start ) && ( sumSNP[ snpNoDis ].genetMap - sumSNP[ cur ].genetMap <= randFrag ); cur-- ) {}
		if( sumSNP[ snpNoDis ].genetMap - sumSNP[ cur ].genetMap > randFrag ) { lowEnd1 = cur + 1; } else { lowEnd1 = cur; }
		randFrag = -log( 1.0 - genrand_real3() ) / lambda;
		for( cur = snpNoDis; ( cur != rgChunk[ chunkNo ].end ) && ( sumSNP[ cur ].genetMap - sumSNP[ snpNoDis ].genetMap <= randFrag ); cur++ ) {}
		if( sumSNP[ cur ].genetMap - sumSNP[ snpNoDis ].genetMap > randFrag ) { highEnd1 = cur - 1; } else { highEnd1 = cur; }
		for( j = lowEnd1; j <= highEnd1; ++ j ){
			gTMRCA = sumSNP[ j ].gType[ 2 ][ 0 ];					// Genotype of MRCA.
			gTSt = sumSNP[ j ].gType[ pool ][ stNo[ i ] ];			// Genotype of subject to be inserted the fragment.
			if( ( gTMRCA == '1' && gTSt == '2' ) || ( gTMRCA == '2' && gTSt == '1' ) ) { sumSNP[ j ].gType[ pool ][ stNo[ i ] ] = gTMRCA; }			
		}
		if ( pool == 0 ) {
		//	fprintf( stderr, "Inserted in one strand PtPool %d: %ld, %ld\n", stNo[ i ], lowEnd1, highEnd1 );
			fprintf( stdout, "Inserted in one strand PtPool %d: %ld, %ld\n", stNo[ i ], lowEnd1, highEnd1 );
		} else {
		//	fprintf( stderr, "Inserted in one strand CtPool %d: %ld, %ld\n", stNo[ i ], lowEnd1, highEnd1 );
			fprintf( stdout, "Inserted in one strand CtPool %d: %ld, %ld\n", stNo[ i ], lowEnd1, highEnd1 );
		}
	}
	free( fPt ); free( stNo );
}

static VALUE wrap_getPossibleRange( VALUE self, VALUE radius, VALUE snp ){
	double r;
	long index, i;
	volatile VALUE	ar;
	
	r = rb_num2dbl( radius );
	index = NUM2LONG( snp );
	ar = rb_ary_new();
	i = index;
	while ( ( i >= 0 ) && ( sumSNP[ i ].chromosome == sumSNP[ index ].chromosome ) && ( sumSNP[ i ].genetMap > sumSNP[ index ].genetMap - r ) ) {
		i--;
	}
	rb_ary_push( ar, LONG2FIX( i ) );
	i = index;
	while ( ( i < noSNP ) && ( sumSNP[ i ].chromosome == sumSNP[ index ].chromosome ) && ( sumSNP[ i ].genetMap < sumSNP[ index ].genetMap + r ) ) {
		i++;
	}
	rb_ary_push( ar, LONG2FIX( i ) );
	return ar;	
}

void Init_HGeneC( void ) {
	volatile VALUE module;
	module = rb_define_module( "HGeneC" );
	
	rb_define_method( module, "initializeCRoutines_C", wrap_initializeCRoutines_C, 5 );
	rb_define_method( module, "readSumSNPData_C", wrap_readSumSNPData_C, 2 );
	rb_define_method( module, "readGenotypingData_C", wrap_readGenotypingData_C, 4 );
	rb_define_method( module, "getSumSNPMem_C", wrap_getSumSNPMem_C, 1 );
	rb_define_method( module, "sumSNPClear", wrap_sumSNPClear, 0 );
	rb_define_method( module, "setupRgChunkAndRgChromosome_C", wrap_setupRgChunkAndRgChromosome_C, 0 );
	rb_define_method( module, "correctErrorHM_C", wrap_correctErrorHM_C, 1 );
	rb_define_method( module, "obtainRHS_C", wrap_obtainRHS_C, 3 );
	rb_define_method( module, "isHMandHHSameInSelectedPt", wrap_isHMandHHSameInSelectedPt, 3 );
	rb_define_method( module, "countOfRCHHInCtPool", wrap_countOfRCHHInCtPool, 3 );
	rb_define_method( module, "takeOverlap_C2", wrap_takeOverlap_C2, 2 );
	rb_define_method( module, "significantQuadriadRegion_C", wrap_significantQuadriadRegion_C, 6 );
	rb_define_method( module, "insertGenotype", wrap_insertGenotype, 4 );
	rb_define_method( module, "performShuffle_C", wrap_performShuffle_C, 1 );
	rb_define_method( module, "saveGenotype", wrap_saveGenotype, 0 );
	rb_define_method( module, "restoreGenotype", wrap_restoreGenotype, 0 );
	rb_define_method( module, "getPossibleRange", wrap_getPossibleRange, 2 );
	rb_define_method( module, "resetNoSNP", wrap_resetNoSNP, 1 );
}

/* For Simulation */
static VALUE wrap_performShuffle_C( VALUE self, VALUE stNo ){
	char **gType;
	long	i, j;
	int	no;
	
	gType = (char**)xcalloc( noSNP, sizeof(char *) );
	for( i = 0; i < noSNP; ++i ){ gType[ i ] = (char*)xcalloc( noPtPool + noCtPool, sizeof(char) ); }
	for( j = 0; j < noPtPool + noCtPool; ++j ){
		no = NUM2INT( *( RARRAY_PTR( stNo ) + j ) );
		//fprintf( stderr, "%d ", no );
		fprintf( stdout, "%d ", no );
		for( i = 0; i < noSNP; ++i ){
			gType[ i ][ j ] = ( no < noPtPool )? sumSNP[ i ].gType[ 0 ][ no ]: sumSNP[ i ].gType[ 1 ][ no - noPtPool ];
		}
	}
	for( i = 0; i < noSNP; ++i ){
		for( j = 0; j < noPtPool + noCtPool; ++j ){
			if( j < noPtPool ) sumSNP[ i ].gType[ 0 ][ j ] = gType[ i ][ j ];
			else sumSNP[ i ].gType[ 1 ][ j - noPtPool ] = gType[ i ][ j ];
		}
	}
	for( i = 0; i < noSNP; ++i ){ free( gType[ i ] ); }
	free( gType );
	//fprintf( stderr, "\n" );
	fprintf( stdout, "\n" );
	return Qnil;
}

static VALUE wrap_saveGenotype( VALUE self ){
	long i, j;
	saveSumSNP = ( struct summerizedSNPdata* )xcalloc( noSNP, sizeof( struct summerizedSNPdata ) );
	for( i = 0; i < noSNP; ++i ){
		saveSumSNP[ i ].gType[ enum_PtPool ] = (unsigned char*)xcalloc( noPtPool + 1, sizeof( unsigned char ) );// Must be calloc to zero memory
		saveSumSNP[ i ].gType[ enum_CtPool ] = (unsigned char*)xcalloc( noCtPool + 1, sizeof( unsigned char ) );
		for( j = 0; j < noPtPool; ++j ){
			saveSumSNP[ i ].gType[ enum_PtPool ][ j ] = sumSNP[ i ].gType[ enum_PtPool ][ j ];
		}
		for( j = 0; j < noCtPool; ++j ){
			saveSumSNP[ i ].gType[ enum_CtPool ][ j ] = sumSNP[ i ].gType[ enum_CtPool ][ j ];
		}
	}
	return Qnil;
}

static VALUE wrap_restoreGenotype( VALUE self ){
	long i, j;
	for( i = 0; i < noSNP; ++i ){
		for( j = 0; j < noPtPool; ++j ){
			sumSNP[ i ].gType[ enum_PtPool ][ j ] = saveSumSNP[ i ].gType[ enum_PtPool ][ j ];
		}
		for( j = 0; j < noCtPool; ++j ){
			sumSNP[ i ].gType[ enum_CtPool ][ j ] = saveSumSNP[ i ].gType[ enum_CtPool ][ j ];
		}
	}
	return Qnil;
}

/*
 *  mersenneTwister.c
 *
 *	Reference: M. Matsumoto and T. Nishimura, "Mersenne Twister: A 623-dimensionally equidistributed uniform pseudorandom number generator", ACM Trans. on Modeling and Computer Simulation Vol. 8, No. 1, January pp.3-30 (1998)
 *
 */

#define N 624
#define M 397
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

/* Period parameters */  

static unsigned long mt[N]; /* the array for the state vector  */
static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */

/* initializes mt[N] with a seed */
void init_genrand(unsigned long s)
{
    mt[0]= s & 0xffffffffUL;
    for (mti=1; mti<N; mti++) {
        mt[mti] = 
	    (1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti); 
        /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
        /* In the previous versions, MSBs of the seed affect   */
        /* only MSBs of the array mt[].                        */
        /* 2002/01/09 modified by Makoto Matsumoto             */
        mt[mti] &= 0xffffffffUL;
        /* for >32 bit machines */
    }
}

/* initialize by an array with array-length */
/* init_key is the array for initializing keys */
/* key_length is its length */
/* slight change for C++, 2004/2/26 */
void init_by_array(unsigned long init_key[], int key_length)
{
    int i, j, k;
    init_genrand(19650218UL);
    i=1; j=0;
    k = (N>key_length ? N : key_length);
    for (; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1664525UL))
          + init_key[j] + j; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i++; j++;
        if (i>=N) { mt[0] = mt[N-1]; i=1; }
        if (j>=key_length) j=0;
    }
    for (k=N-1; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 30)) * 1566083941UL))
          - i; /* non linear */
        mt[i] &= 0xffffffffUL; /* for WORDSIZE > 32 machines */
        i++;
        if (i>=N) { mt[0] = mt[N-1]; i=1; }
    }

    mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */ 
}

/* generates a random number on [0,0xffffffff]-interval */
unsigned long genrand_int32(void)
{
    unsigned long y;
    static unsigned long mag01[2]={0x0UL, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)   /* if init_genrand() has not been called, */
            init_genrand(5489UL); /* a default initial seed is used */

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

        mti = 0;
    }
  
    y = mt[mti++];

    /* Tempering */
    y ^= (y >> 11);
    y ^= (y << 7) & 0x9d2c5680UL;
    y ^= (y << 15) & 0xefc60000UL;
    y ^= (y >> 18);

    return y;
}

/* generates a random number on [0,0x7fffffff]-interval */
long genrand_int31(void)
{
    return (long)(genrand_int32()>>1);
}

/* generates a random number on [0,1]-real-interval */
double genrand_real1(void)
{
    return genrand_int32()*(1.0/4294967295.0); 
    /* divided by 2^32-1 */ 
}

/* generates a random number on [0,1)-real-interval */
double genrand_real2(void)
{
    return genrand_int32()*(1.0/4294967296.0); 
    /* divided by 2^32 */
}

/* generates a random number on (0,1)-real-interval */
double genrand_real3(void)
{
    return (((double)genrand_int32()) + 0.5)*(1.0/4294967296.0); 
    /* divided by 2^32 */
}

/* generates a random number on [0,1) with 53-bit resolution*/
double genrand_res53(void) 
{ 
    unsigned long a=genrand_int32()>>5, b=genrand_int32()>>6; 
    return(a*67108864.0+b)*(1.0/9007199254740992.0); 
} 
/* These real versions are due to Isaku Wada, 2002/01/09 added */

