require "HGeneC"
require "FileUtils"

class ReverseLineReader #http://www.clear-code.com/blog/2010/6/23.html
  def initialize(io)
    @io = io
    @io.seek(0, IO::SEEK_END)
    @buffer = ""
    @data = ""
  end

  def each
    separator = $/
    separator_length = separator.length
    while read_to_buffer
      loop do
        index = @buffer.rindex(separator, @buffer.length - 1 - separator_length)
        break if index.nil? or index.zero?
        last_line = @buffer.slice!((index + separator_length)..-1)
        yield(last_line)
      end
    end
    yield(@buffer) unless @buffer.empty?
  end

  private
  BYTES_PER_READ = 4096
  def read
    position = @io.pos
    if position < BYTES_PER_READ
      bytes_per_read = position
    else
      bytes_per_read = BYTES_PER_READ
    end

    if bytes_per_read.zero?
      @data.replace("")
    else
      @io.seek(-bytes_per_read, IO::SEEK_CUR)
      @io.read(bytes_per_read, @data)
      @io.seek(-bytes_per_read, IO::SEEK_CUR)
    end

    @data
  end

  def read_to_buffer
    data = read
    if data.empty?
      false
    else
      @buffer.insert(0, data)
      true
    end
  end
end

class HGeneRb < Object
  include HGeneC

  Mode_HMAndHH = 1
  Mode_HHOnly = 0

  attr_accessor :sumSNPFile
  attr_accessor :annotationFile
  attr_accessor :patientPoolFiles
  attr_accessor :controlPoolFiles
  attr_accessor :targetAreaFile
  attr_accessor :excludedAreaFile
  attr_accessor :outputDirectory
  attr_accessor :suspectedGenotypingErrorRate
  attr_accessor :appliedErrorCorrectionRate
  attr_accessor :rHSCutoff
  attr_accessor :noPtPool
  attr_accessor :noCtPool
  attr_accessor :rgChunk
  attr_accessor :rgChromosome
  attr_accessor :rsIDHash
  attr_accessor :largeGap
  attr_accessor :lastChromosome
  attr_accessor :extraGenotypingFiles
  attr_accessor :noSNP
  
  attr_accessor :noCtAllowedToShareRHS
  attr_accessor :noCtAllowedToShareRCHH
  attr_accessor :maxNoPositiveCombinations
  attr_accessor :minLevelToDrawBirdsEyeFigure
  attr_accessor :maxLevel
  attr_accessor :generationM
  attr_accessor :insertFragmentIBD
  attr_accessor :snpNoCausingDisease
  attr_accessor :noPtInsertedFragmentIBD
  attr_accessor :backgroundPercentageOfIBD
  attr_accessor :omitComparingRCHH
  attr_accessor :omitComparingRHS
  attr_accessor :shuffle
  attr_accessor :rangeStart
  attr_accessor :rangeEnd
  attr_accessor :mereHM
  
  def initialize()
    @noSNP = 0
    @sumSNPFile = ""
    @annotationFile = ""
    @patientPoolFiles = []
    @controlPoolFiles = []
    @targetAreaFile = ""
    @excludedAreaFile = ""
    @outputDirectory = ""                       #outputDir
    @suspectedGenotypingErrorRate = 0.0
    @appliedErrorCorrectionRate = 0.0
    @rHSCutoff = 1.2                            #rHSCutoff
    @noPtPool = 0; @noCtPool = 0; @noExtra = 0
    @rgChunk = []
    @rgChromosome = []
    @rsIDHash = Hash.new
    @largeGap = 300000
    @lastChromosome = 22
    @extraGenotypingFiles = []
    
    @noCtAllowedToShareRHS = 0            #nCtShareRHS
    @noCtAllowedToShareRCHH = 0           #nCtShareRCHH
    @maxNoPositiveCombinations = 100000   #maNPosComb
    @minLevelToDrawBirdsEyeFigure = 5     #miLevelFig
    @maxLevel = 10                        #maxLevel
    @generationM = 0                      #genM
    @insertFragmentIBD = false            #insIBD
    @snpNoCausingDisease = -1             #sNDis
    @noPtInsertedFragmentIBD = 0          #nPtInsIBD
    @backgroundPercentageOfIBD = 0        #bkgd
    @omitComparingRHS = false             #omitCompRHS
    @omitComparingRCHH = false            #omitCompRCHH
    @shuffle = false                      #shuffle
    @hhON = true                          #hhON
    @rangeStart = 0                       #rgStart
    @rangeEnd = 10                        #rgEnd
    @mereHM = false                       #mereHM
  end
  
  def readOptionsAndAnalysisFile( ar )
    ar.each{ | a |
      if a !~ /^-/ && FileTest.exist?( a ) then readAnalysisFile( a ) end
    }
    ar.each{ | a |
      if a =~ /-nCtShareRHS:(\d+)/ then @noCtAllowedToShareRHS = $1.to_i end
      if a =~ /-nCtShareRCHH:(\d+)/ then @noCtAllowedToShareRCHH = $1.to_i end
      if a =~ /-maNPosComb:(\d+)/ then @maxNoPositiveCombinations = $1.to_i end
      if a =~ /-miLevelFig:(\d+)/ then @minLevelToDrawBirdsEyeFigure = $1.to_i end
      if a =~ /-maxLevel:(\d+)/ then @maxLevel = $1.to_i end
      if a =~ /-genM:(\d+)/ then @generationM = $1.to_i end
      if a =~ /-sNDis:(\d+)/ then @snpNoCausingDisease = $1.to_i end
      if a =~ /-nPtInsIBD:(\d+)/ then @noPtInsertedFragmentIBD = $1.to_i; @insertFragmentIBD = true end
      if a =~ /-bkgd:(\d+)/ then @backgroundPercentageOfIBD = $1.to_i; @insertFragmentIBD = true end
      if a =~ /-insIBD:YES/ then @insertFragmentIBD = true end
      if a =~ /-insIBD:NO/ then @insertFragmentIBD = false end
      if a =~ /-omitCompRHS:YES/ then @omitComparingRHS = true end
      if a =~ /-omitCompRHS:NO/ then @omitComparingRHS = false end
      if a =~ /-omitCompRCHH:YES/ then @omitComparingRCHH = true end
      if a =~ /-omitCompRCHH:NO/ then @omitComparingRCHH = false end
      if a =~ /-mereHM:YES/ then @mereHM = true end
      if a =~ /-mereHM:NO/ then @mereHM = false end
      if a =~ /-shuffle:YES/ then @shuffle = true end
      if a =~ /-shuffle:NO/ then @shuffle = false end
      if a =~ /-rHSCutoff:([\d.]+)/ then @rHSCutoff = $1.to_f end
      if a =~ /-hhON:YES/ then @hhON = true end
      if a =~ /-hhON:NO/ then @hhON = false end
      if a =~ /-rgStart:(\d+)/ then @rangeStart = $1.to_i end
      if a =~ /-rgEnd:(\d+)/ then @rangeEnd = $1.to_i end
      if a =~ /-outputDir:(.+)/ then
        dirName = $1.delete('"')
        if FileTest.exist?( $1 ) == false then FileUtils.mkdir_p( $1 ) end
        @outputDirectory = $1
      end
    }
    if ( @noPtInsertedFragmentIBD != 0 || @backgroundPercentageOfIBD != 0 ) && @insertFragmentIBD == false then STDERR.puts "Set -insIBD correctly!!!" end
    if ( @noPtInsertedFragmentIBD != 0 || @backgroundPercentageOfIBD != 0 ) && @snpNoCausingDisease == -1 then STDERR.puts "Set -sNDis!!!"; exit end
    if ( @noPtInsertedFragmentIBD != 0 || @backgroundPercentageOfIBD != 0 ) && @generationM == 0 then STDERR.puts "Set -genM!!!"; exit end
    [ @sumSNPFile, @annotationFile, @patientPoolFiles, @controlPoolFiles, @targetAreaFile, 
      @excludedAreaFile, @outputDirectory, @suspectedGenotypingErrorRate, @rHSCutoff, @largeGap,
      @lastChromosome, @extraGenotypingFiles, @noCtAllowedToShareRCHH, @mereHM, @maxNoPositiveCombinations,
      @minLevelToDrawBirdsEyeFigure, @maxLevel, @generationM, @snpNoCausingDisease, @noPtInsertedFragmentIBD,
      @backgroundPercentageOfIBD, @insertFragmentIBD, @omitComparingRHS, @omitComparingRCHH ]  #For testing
  end
  
  def readAnalysisFile( fileName )
    require "stringio"
    nAr = [ "sumSNPFile", "annotationFile", "patientPoolFiles", "controlPoolFiles", "targetAreaFile", 
          "excludedAreaFile", "outputDirectory", "suspectedGenotypingErrorRate", "rHSCutoff", "largeGap",
          "lastChromosome", "extraGenotypingFiles", "noCtAllowedToShareRHS", "noCtAllowedToShareRCHH",
          "maxNoPositiveCombinations", "minLevelToDrawBirdsEyeFigure", "maxLevel" ]
    io = StringIO.new
    line = File.read( fileName ) + "\n::"
    [ /(::sumSNP file.+?)::/m, /(::annotation file.+?)::/m, /(::patient pool.+?)::/m, /(::control pool.+?)::/m,
      /(::target area file.+?)::/m, /(::excluded area file.+?)::/m, /(::output directory.+?)::/m,
      /(::suspected genotyping error rate.+?)::/m, /(::RHS cutoff.+?)::/m, /(::large gap.+?)::/m,
      /(::last chromosome.+?)::/m, /(::extra genotyping files.+?)::/m, /(::number of control allowed to share rhs.+?)::/m, /(::number of control allowed to share rchh.+?)::/m,
      /(::max number of positive combinations.+?)::/m, /(::min level to draw figure.+?)::/m, /(::max level.+?)::/m
       ].each_with_index do | str, i |
      st = ""; tAr = []
      if line =~ str
        io.string = $1
        line1 = io.gets
        while line1 = io.gets
          line1.chomp!.strip =~ /(.+)/
          case i
          when 0, 1, 4, 5, 6
            if $1 != nil then self.send( nAr[ i ] + "=", $1 ); break; end
          when 7, 8
            if $1 != nil then self.send( nAr[ i ] + "=", $1.to_f ); break; end
          when 2, 3, 11
            if $1 != nil then self.send( nAr[ i ] + "=", self.send( nAr[ i ] ) << $1 ) end
          when 9, 10, 12, 13, 14, 15, 16
            if $1 != nil then self.send( nAr[ i ] + "=", $1.to_i ); break; end
          end
        end
      end
    end
    initializeCRoutines
    [ @patientPoolFiles, @controlPoolFiles ].each do | ar |
      ar.each do | file |
        if file != "" && File.exists?( file ) == false
          STDERR.puts "#{ file } does not exist.  Quitting program...."
          exit
        end
      end
    end
    [ @sumSNPFile, @annotationFile, @targetAreaFile, @excludedAreaFile ].each do | file |
      if file != "" && File.exists?( file ) == false
        STDERR.puts "#{ file } does not exist.  Quitting program...."
        exit
      end
    end
    nil
  end
  
  #Setup @noSNP, @noPtPool, @noCtPool, @noExtra and allocate memory in C world.
  def initializeCRoutines
    if @patientPoolFiles != []
      File.open( @annotationFile, "r" ) do | file |
        ReverseLineReader.new( file ).each do | line |
          @noSNP = line.chomp!.split( "\t" )[ 5 ].to_i + 1
          break
        end
      end
      @noPtPool = @patientPoolFiles.length
      @noCtPool = @controlPoolFiles.length
      @noExtra = @extraGenotypingFiles.length
    elsif @sumSNPFile != ""
      File.open( @sumSNPFile, "r" ) do | file |
        file.gets =~ /TotalSNPNumber=(\d+)\tNumberOfSubjectsInPatientPool=(\d+)\tNumberOfSubjectsInControlPool=(\d+)\tNumberOfExtraSubjects=(\d+)\tErrorCorrection_HM=([.\d]+)/
        @noSNP = $1.to_i; @noPtPool = $2.to_i; @noCtPool = $3.to_i; @noExtra = $4.to_i; @appliedErrorCorrectionRate = $5.to_f;
      end
    else
      STDERR.puts "Neither sumSNP file nor patient files are set."
      exit
    end
    initializeCRoutines_C( @noSNP, @noPtPool, @noCtPool, @noExtra, @largeGap )
  end

  def readAnnotationFile
    chrCol = 0; posCol = 1; aAlleleFreqCol = 2; rsIDAnnotCol = 3; genetMapAnnotCol = 4; patientGTypeCol = 5; controlGTypeCol = 6; extraGTypeCol = 7
    count = 0
    STDERR.puts "Reading annotation file... "
    File.foreach( self.annotationFile ) do | line |
      line =~ /^(.*?)\t(.*?)\t(.*?)\t(.*?)\t(.*?)\t/
      tAr = [ $1, $2, $3, $4, $5 ]
      if @lastChromosome < $1.to_i then next end
      @rsIDHash[ $4 ] = count
      readSumSNPData_C( tAr, count )
      count += 1
      if count % 100000 == 0 then STDERR.print "#{ count } " end
    end
    STDERR.puts ""
    @noSNP = count  #noSNP is reassigned
    resetNoSNP count  #also reset the variable noSNP in C routine.
    @rgChunk, @rgChromosome = setupRgChunkAndRgChromosome_C
    [ rgChunk ] #for testing
  end

  def readSumSNPFile
    chrCol = 0; posCol = 1; aAlleleFreqCol = 2; rsIDAnnotCol = 3; genetMapAnnotCol = 4; patientGTypeCol = 5; controlGTypeCol = 6; extraGTypeCol = 7
    count = 0
    STDERR.puts "Reading sumSNP file..."
    File.foreach( @sumSNPFile ) do | line |
      line =~ /^(.*?)\t(.*?)\t(.*?)\t(.*?)\t(.*?)\t(.*?)\t(.*?)\t(.*?)\t/
      tAr = [ $1, $2, $3, $4, $5, $6, $7, $8 ]
      if @lastChromosome < $1.to_i then next end
      @rsIDHash[ $4 ] = count
      readSumSNPData_C( tAr, count )
      count += 1
      if count % 100000 == 0 then STDERR.print "#{ count } " end
    end
    STDERR.puts ""
    @noSNP = count  #noSNP is reassigned
    @rgChunk, @rgChromosome = setupRgChunkAndRgChromosome_C
    [ rgChunk ] #for testing
  end
  
  def readGenotypingData( pool )
    if pool == 0 then gFileAr = patientPoolFiles elsif pool == 1 then gFileAr = controlPoolFiles else gFileAr = extraGenotypingFiles end
    gFileAr.each_with_index do | file, i |
      nValidRSID = 0; nInvalidRSID = 0
      STDERR.print( "Processing #{ file }...")
      File.foreach( file ) do | line |
        line =~ /(rs\d+)/
        index = @rsIDHash[ $1 ]
        line =~ /\t([AB][AB])/
        gt = $1
        if index == nil then nInvalidRSID += 1; next; else nValidRSID += 1 end
        case gt when "AA" then gType = "1" when "BB" then gType = "2" when "AB" then gType = "3" else gType = "0" end
        readGenotypingData_C( index, i, gType[ 0 ], pool )
      end
      aa, bb, ab = readGenotypingData_C( 0, 0, 0, -1 )
      STDERR.print( " Valid SNP_ID:#{ nValidRSID }, Invalid SNP_ID:#{ nInvalidRSID }, AA:#{ aa }, BB:#{ bb }, AB:#{ ab }, NoCall:#{ @noSNP - ( aa + bb + ab ) }\n" )
    end
  end
  
  def performShuffle
    performShuffle_C( Array.new( @noPtPool + @noCtPool ){ | i | [ i, rand ] }.sort{ | a, b | a[ 1 ] <=> b[ 1 ] }.map{ | a | a[ 0 ] } )
  end
  
  def correctErrorHM
    if @appliedErrorCorrectionRate != 0.0
      if suspectedGenotypingErrorRate != 0.0
        STDERR.puts "****Genotyping error had already been corrected and is not done in the current session****"
      end
    else
      if suspectedGenotypingErrorRate != 0.0
        @appliedErrorCorrectionRate = correctErrorHM_C( suspectedGenotypingErrorRate )
      end
    end
  end
  
  def outputSumSNPFile( outputFileName )
    STDERR.puts "Outputting sumSNP file as #{ outputFileName }... "
    io = open( outputFileName, "w" )
    add = "ErrorCorrection_HM=%0.4f" % @appliedErrorCorrectionRate
    for i in 0...@noSNP do
      ar = getSumSNPMem_C( i )
      if i == 0
        io.puts "#{ ar.join( "\t" ) }\tTotalSNPNumber=#{ @noSNP }\tNumberOfSubjectsInPatientPool=#{ @noPtPool }\tNumberOfSubjectsInControlPool=#{ @noCtPool }\tNumberOfExtraSubjects=#{ @noExtra }\t#{ add }"
      else
        io.puts ar.join( "\t" )
      end
    end
    io.close
  end
  
  def outputPS_BW( resultAr, outfile )
    if resultAr.length == 0 then return end
    io = open( outfile, "w" )
    io.puts "%!PS-Adobe-3.1 EPSF-3.0
    28.3 28.3 translate
    /str 20 string def
    /Helvetica findfont 12 scalefont setfont
    1 1 22 { dup -8 exch 1 sub 0.75 mul 1.15 add 18 mul moveto (22) stringwidth pop 2 div neg 0 rmoveto str cvs dup stringwidth pop 2 div neg 0 rmoveto show } for
    2.83 2.83 scale
    0 setgray
    0.5 setlinewidth
    25 0 moveto 29 -2.5 lineto 29 -1.75 lineto 82 -1.75 lineto 82 1.75 lineto 29 1.75 lineto 29 2.5 lineto closepath stroke
    150 0 moveto 146 -2.5 lineto 146 -1.75 lineto 92 -1.75 lineto 92 1.75 lineto 146 1.75 lineto 146 2.5 lineto closepath stroke
    /Helvetica findfont 5 scalefont setfont
    10 -1.5 moveto (p-ter) show
    165 -1.5 moveto (q-ter) dup stringwidth pop neg 0 rmoveto show
    0.353357 0.353357 scale
    0.02 18 scale
    0 setlinewidth
    0.9 setgray"

    @rgChunk.each{ | a |
      ar = getSumSNPMem_C( a[ 0 ] )
      io.puts "#{ ar[ 0 ] }\t#{ ar[ 1 ].to_f / 10000.0 }\texch 1 sub 0.75 mul 1 add moveto 0 0.5 rlineto"
      ar = getSumSNPMem_C( a[ 1 ] )
      io.puts "#{ ar[ 0 ] }\t#{ ar[ 1 ].to_f / 10000.0 }\texch 1 sub 0.75 mul 1.5 add lineto 0 -0.5 rlineto closepath fill"
    }
    
    io.puts "0 setgray"
    
    resultAr.each{ | a |
      n = a[ 0 ]
      ar = getSumSNPMem_C( n )
      io.puts "#{ ar[ 0 ] }\t#{ ar[ 1 ].to_f / 10000.0 }\texch 1 sub 0.75 mul 1 add moveto 0 0.5 rlineto\t%%#{ ar[ 3 ] }\tstart\t#{ n }"
      n = a[ 1 ]
      ar = getSumSNPMem_C( n )
      io.puts "#{ ar[ 0 ] }\t#{ ar[ 1 ].to_f / 10000.0 }\texch 1 sub 0.75 mul 1.5 add lineto 0 -0.5 rlineto closepath fill\t%%#{ ar[ 3 ] }\tend\t#{ n }"
    }
    io.close
  end

  def outputText_BW( resultAr, outfile )
    if resultAr.length == 0 then return end
    io = open( outfile, "w" )
    resultAr.each{ | a |
      n = a[ 0 ]
      ar = getSumSNPMem_C( n )
      io.puts "#{ ar[ 0 ] }\t#{ ar[ 1 ] }\t#{ ar[ 4 ] }\t#{ ar[ 3 ] }\tstart"
      n = a[ 1 ]
      ar = getSumSNPMem_C( n )
      io.puts "#{ ar[ 0 ] }\t#{ ar[ 1 ] }\t#{ ar[ 4 ] }\t#{ ar[ 3 ] }\tend"
    }
    io.close
  end

  def rgChunkNo( index )
    for i in 0...@rgChunk.length
      if @rgChunk[ i ][ 0 ] <= index && index <= @rgChunk[ i ][ 1 ] then break end
    end
    i
  end

  def hm_on_hh_PtPoolSubgroup_vs_CtPool( numberOfPtInSubgroup, thresholdCtStNumberHM, thresholdCtStNumberHH )
    if numberOfPtInSubgroup <= 1 then STDERR.puts "2nd argument of hm_on_hh_PtPoolSubgroup_vs_CtPool should be 2 =<" end
    number = nCr( @noPtPool, numberOfPtInSubgroup )
    STDERR.puts "Processing numberOfPtInSubgroup = #{ numberOfPtInSubgroup }: #{ number } combinations..."
    count = 0
    $rgHashAr[ numberOfPtInSubgroup - 1 ].each do | k, v |
      ptNumAr1 = k.to_s.scan( /\d{3}/ ).map{ | a | a.to_i } #ptNumAr1 is like [ 0, 2, 3, 4 ] when numberOfPtInSubgroup == 5
      for ptNum2 in ( ptNumAr1[ -1 ] + 1 )...@noPtPool do   #ptNum2 is like 5, 6, 7, ...59
        if numberOfPtInSubgroup == 2                        #if ptNumAr1 is like [ 0, 2 ]
          overlapArAr = [ v, $rgHashAr[ 1 ][ ( "_%03d" % ptNum2 ).to_sym ] ]  #ovelapArAr = [ v, [ [st, ed], [st, ed], ...] ]
          ar = takeOverlap_C2( overlapArAr, 2 )
          if @mereHM == false then                              
            ar = significantQuadriadRegion_C( ar, [ ptNumAr1[ 0 ], ptNum2 ], thresholdCtStNumberHM, thresholdCtStNumberHH, @omitComparingRHS, @omitComparingRCHH )
          end
          ar = ar.map{ | a | [ a[ 0 ], a [ 1 ] ] }  #ar is like [ [st, ed], [st, ed], ... ]
        else
          flag = 0
          if numberOfPtInSubgroup == 3 then
            if $rgHashAr[ 2 ][ ( ( "_%03d" % ptNumAr1[ 0 ] ) + ( "_%03d" % ptNumAr1[ 1 ] ) ).to_sym ] == nil ||
               $rgHashAr[ 2 ][ ( ( "_%03d" % ptNumAr1[ 0 ] ) + ( "_%03d" % ptNum2 ) ).to_sym ] == nil ||
               $rgHashAr[ 2 ][ ( ( "_%03d" % ptNumAr1[ 1 ] ) + ( "_%03d" % ptNum2 ) ).to_sym ] == nil then
               flag = 1
            end
          elsif numberOfPtInSubgroup > 3 then
            for i in 0...( ptNumAr1.length - 1 ) do
              if $rgHashAr[ 3 ][ ( ( "_%03d" % ptNumAr1[ i ] ) + ( "_%03d" % ptNumAr1[ i + 1 ] ) + ( "_%03d" % ptNum2 ) ).to_sym ] == nil then flag = 1; break; end
            end
          end
          if flag == 1 then
            next
          else
            overlapArAr = [ v ]
            ptNumAr1.each do | ptNum1 |
              overlapArAr << $rgHashAr[ 2 ][ ( ( "_%03d" % ptNum1 ) + ( "_%03d" % ptNum2 ) ).to_sym ]
            end
            ar = takeOverlap_C2( overlapArAr, numberOfPtInSubgroup )
          end
        end
        #ar is like [ [st, ed], [st, ed], ... ]
        if( ar.length != 0 ) then $rgHashAr[ numberOfPtInSubgroup ][ ( k.to_s + ( "_%03d" % ptNum2 ) ).to_sym ] = ar; count += 1 end
        #if number of combinations > 100000
        if count >= @maxNoPositiveCombinations then STDERR.puts "So many positive combinations (>=100000).  Interrupt collecting combinations"; break end
      end
      if count >= @maxNoPositiveCombinations then break end
    end
    STDERR.puts "Finished numberOfPtInSubgroup = #{ numberOfPtInSubgroup }: #{ count } positive signals in #{ number } combinations."
    STDOUT.puts "Finished numberOfPtInSubgroup = #{ numberOfPtInSubgroup }: #{ count } positive signals in #{ number } combinations."
    count
  end

  #Required variable---$rgHashAr
  #TakeOverlap_C2   required variables in C:noSNP, noChunk, rgChunk[]
  def horizontalFigure( level, outfile )
    ar = []
    $rgHashAr[ level ].each do | key, value |
      ar << value
    end
    ar = takeOverlap_C2( ar, -2 )  #multicolor mode
    if ar.length == 0 then return nil end
    max = ar.max{ | a, b | a[ 1 ] <=> b[ 1 ] }
    if max == nil || max[ 1 ] == 0 then return nil end  # If the figure is blank, do not draw it.
    xFact = 1.0; yFact = 1.0
    io = open( outfile, "w" )
    io.puts "%!PS-Adobe-3.1 EPSF-3.0
    72 72 translate 0.5 setlinewidth
    (Times-Roman) findfont 16 scalefont setfont
    15 190 moveto (Level #{ level }) show
    (Times-Roman) findfont 12 scalefont setfont
    -15 0 moveto -10 0 lineto -10 180 lineto -15 180 lineto stroke
    -20 -4 moveto (0) stringwidth pop neg 0 rmoveto (0) show
    -20 176 moveto (#{ max[ 1 ] }) stringwidth pop neg 0 rmoveto (#{ max[ 1 ] }) show
    gsave -18 86 moveto -18 86 translate 90 rotate
    (Number of subgroups) stringwidth pop neg 2 div 0 rmoveto (Number of subgroups) show
    grestore
    0 setlinewidth"

    xFact = 0.0005; yFact = 18

    gray = 0.9
    @rgChromosome.each_with_index{ | a, i |
      ( gray == 0.9 )? gray = 0.8: gray = 0.9
      io.puts "#{ gray } setgray"
      io.puts "#{ a[ 0 ].to_f * xFact } 0 moveto #{ a[ 0 ].to_f * xFact } #{ -0.8 * yFact } lineto"
      io.puts "#{ a[ 1 ].to_f * xFact } #{ -0.8 * yFact } lineto #{ a[ 1 ].to_f * xFact } 0 lineto closepath fill"
    }
    io.puts "0 setgray 0.5 setlinewidth"
    yFact *= ( 10.0 / max[ 1 ] )
    ar.each_with_index{ | a, i |
      if i != ar.length - 1
        io.puts "#{ ar[ i ][ 0 ].to_f * xFact } #{ ar[ i ][ 1 ].to_f * yFact } moveto #{ ar[ i + 1 ][ 0 ].to_f * xFact } #{ ar[ i + 1 ][ 1 ].to_f * yFact } lineto stroke\t%%#{ ar[ i ][ 0 ] },  #{ ar[ i ][ 1 ] } to #{ ar[ i + 1 ][ 0 ] }, #{ ar[ i + 1 ][ 1 ] }\n"
      end
    }
    yFact *= ( max[ 1 ] / 10.0 )
    @rgChromosome.each_with_index{ | a, i |
      middle = ( a[ 0 ].to_f + a[ 1 ].to_f ) / 2.0
      if i < 14
        io.puts "#{ middle * xFact } #{ -0.65 * yFact } moveto gsave 0 setgray (#{ i + 1 }) stringwidth pop 2 div neg 0 rmoveto (#{ i + 1 }) show grestore"
      else
        io.puts "#{ middle * xFact } #{ -0.65 * yFact } moveto gsave 0 setgray #{ 1.0 - ( i - 13 ) * 0.7 / 9 } 1 scale (#{ i + 1 }) stringwidth pop 2 div neg 0 rmoveto (#{ i + 1 }) show grestore"
      end
    }
    io.close
  end
  
  def nCr( n, r )
    a, b = r, n-r
    a, b = b, a if a < b  # a is the larger
    numer = ( a + 1..n ).inject( 1 ) { | t, v | t * v }  # n!/r!
    denom = ( 2..b ).inject( 1 ) { | t,  v | t * v }    # (n-r)!
    numer / denom
  end

  def outputReport( io )
    io.print "sumSNPFile= "; io.print @sumSNPFile; io.print ", "
    io.print "annotationFile= "; io.print @annotationFile; io.print ", "
    io.print "patientPoolFiles= "; io.print @patientPoolFiles.inspect; io.print ", "
    io.print "controlPoolFiles= "; io.print @controlPoolFiles.inspect; io.print ", "
    io.print "targetAreaFile= "; io.print @targetAreaFile; io.print ", "
    io.print "excludedAreaFile= "; io.print @excludedAreaFile; io.print ", "
    io.print "outputDirectory= "; io.print @outputDirectory; io.print ", "
    io.print "suspectedGenotypingErrorRate= "; io.print @suspectedGenotypingErrorRate; io.print ", "
    io.print "rHSCutoff= "; io.print @rHSCutoff; io.print ", "
    io.print "noPtPool= "; io.print @noPtPool; io.print ", "
    io.print "noCtPool= "; io.print @noCtPool; io.print ", "
    io.print "largeGap= "; io.print @largeGap; io.print ", "
    io.print "lastChromosome= "; io.print @lastChromosome; io.print ", "
    io.print "extraGenotypingFiles= "; io.print @extraGenotypingFiles.inspect; io.print ", "
    io.print "noCtAllowedToShareRHS= "; io.print @noCtAllowedToShareRHS; io.print ", "
    io.print "noCtAllowedToShareRCHH= "; io.print @noCtAllowedToShareRCHH; io.print ", "
    io.print "maxNoPositiveCombinations= "; io.print @maxNoPositiveCombinations; io.print ", "
    io.print "minLevelToDrawBirdsEyeFigure= "; io.print @minLevelToDrawBirdsEyeFigure; io.print ", "
    io.print "maxLevel= "; io.print @maxLevel; io.print ", "
    io.print "generationM= "; io.print @generationM; io.print ", "
    io.print "insertFragmentIBD= "; io.print @insertFragmentIBD; io.print ", "
    io.print "snpNoCausingDisease= "; io.print @snpNoCausingDisease; io.print ", "
    io.print "noPtInsertedFragmentIBD= "; io.print @noPtInsertedFragmentIBD; io.print ", "
    io.print "backgroundPercentageOfIBD= "; io.print @backgroundPercentageOfIBD; io.print ", "
    io.print "omitComparingRHS= "; io.print @omitComparingRHS; io.print ", "
    io.print "omitComparingRCHH= "; io.print @omitComparingRCHH; io.print ", "
    io.print "shuffle= "; io.print @shuffle; io.print "\n\n"
  end
end

